/* dhcpd.c
 *
 * udhcp DHCP client
 *
 * Russ Dill <Russ.Dill@asu.edu> July 2001
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/file.h>
#include <unistd.h>
#include <getopt.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <time.h>
#include <string.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <errno.h>

#include "dhcpd.h"
#include "dhcpc.h"
#include "options.h"
#include "clientpacket.h"
#include "script.h"
#include "socket.h"
#include "debug.h"

#include "proto.h"

#define LISTEN_NONE 0
#define LISTEN_KERNEL 1
#define LISTEN_RAW 2
static int listen_mode = LISTEN_RAW;
static int ifindex=-1;
static unsigned char arp[6];

static char ETHNAME[128];
static char SCRIPTFILE[128];

static unsigned long int ethaddr=0;

unsigned long getipaddr(char *device)
{
    int fd;
    struct ifreq ifr;
    struct sockaddr_in *sin;

    if((fd=socket(AF_INET,SOCK_DGRAM,0))<0)
        return 0;
    memset((char *)&ifr,0,sizeof(ifr));
    strncpy(ifr.ifr_name,device,sizeof(ifr.ifr_name));
    sin = (struct sockaddr_in *)&ifr.ifr_addr;
    sin->sin_family = AF_INET;
    if(ioctl(fd,SIOCGIFADDR,(char *)&ifr)==-1)
    {
        close(fd);
        return 0;
    }
    close(fd);
    return sin->sin_addr.s_addr;
}

unsigned long int LoadEThIP() {

	ethaddr = getipaddr(ETHNAME);
	return ethaddr;
}

int configure_eth(char *interf,char *script) {

	int fd=-1;
	struct ifreq ifr;
	ETHNAME[0]=0;
	if  ( interf) strcpy(ETHNAME,interf);
	SCRIPTFILE[0] = 0;
	if ( script) strcpy(SCRIPTFILE,script);

	if ((fd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) >= 0) {
		strcpy(ifr.ifr_name, interf);
		if (ioctl(fd, SIOCGIFINDEX, &ifr) == 0) {
 			ifindex=ifr.ifr_ifindex;
			DEBUG(LOG_INFO, "adapter index %d", ifr.ifr_ifindex);
		} else {
			LOG(LOG_ERR, "SIOCGIFINDEX failed! %s", strerror(errno));
			return 0 ;
		}
		if (ioctl(fd, SIOCGIFHWADDR, &ifr) == 0) {
			memcpy(arp, ifr.ifr_hwaddr.sa_data, 6);
			DEBUG(LOG_INFO, "adapter hardware address %02x:%02x:%02x:%02x:%02x:%02x",arp[0], arp[1], arp[2], arp[3], arp[4], arp[5]);
		} else {
			LOG(LOG_ERR, "SIOCGIFHWADDR failed! %s", strerror(errno));
			return 0 ;
		}
	} else {
		LOG(LOG_ERR, "socket failed! %s", strerror(errno));
		return 0 ;
	}
	close(fd);
	fd = -1;
	return 1;

}

int RequestIPFromDHCP(char*inte,char *clientId,unsigned long int  *ipac,unsigned long int *ipserver,long int *lease, char **msdns, char **mswins, int istoconfig) {
       
	int fd=-1,retval;
	struct timeval tv;
	unsigned long xid = 0;
	char *temp,*message;
	fd_set rfds;
	struct dhcpMessage packet;
	struct in_addr temp_addr;
	int state;
	unsigned long requested_ip=0;
	unsigned long server_addr;
	unsigned long timeout=0;
	int packet_num=0;

	state = INIT_SELECTING;
	for (;;) {
		if (fd > 0) {
			close(fd);
			fd = -1;
		}
		
		if (listen_mode == LISTEN_KERNEL) {
			if ((fd = listen_socket(INADDR_ANY, CLIENT_PORT, ETHNAME)) < 0) {
				LOG(LOG_ERR, "couldn't create server socket -- au revoir");
				return 0;
			}			
		} else if (listen_mode == LISTEN_RAW) {
			if ((fd = raw_socket(ifindex)) < 0) {
				LOG(LOG_ERR, "couldn't create raw socket -- au revoir");
				return 0;
			}			
		} else fd = -1;

		tv.tv_sec = timeout - time(0);
		tv.tv_usec = 0;
		FD_ZERO(&rfds);
		if (listen_mode) FD_SET(fd, &rfds);
		
		if (tv.tv_sec > 0) {
			retval = select(fd + 1, &rfds, NULL, NULL, &tv);
		} else retval = 0; /* If we already timed out, fall through */
		
		if (retval == 0) {
			/* timeout dropped to zero */
			switch (state) {
			case INIT_SELECTING:
				if (packet_num < 3) {
					if (packet_num == 0) xid = random_xid();
					send_discover(xid, requested_ip,ifindex,arp,clientId); /* broadcast */		
					timeout = time(0) + ((packet_num == 2) ? 10 : 2);
					packet_num++;
				} else {
						return 0 ;
				}
				break;
			case REQUESTING:
				if (packet_num < 3) {
					/* send request packet */
					send_selecting(xid, server_addr, requested_ip,ifindex,arp,clientId); /* broadcast */	
					timeout = time(0) + ((packet_num == 2) ? 10 : 2);
					packet_num++;
				} else {
					return 0;
				}
				break;
			}
		} else if (retval > 0 && listen_mode != LISTEN_NONE && FD_ISSET(fd, &rfds)) {
			/* a packet is ready, read it */
			
			if (listen_mode == LISTEN_KERNEL) {
				if (get_packet(&packet, fd) < 0) continue;
			} else {
				if (get_raw_packet(&packet, fd) < 0) continue;
			} 
			
			if (packet.xid != xid) {
				DEBUG(LOG_INFO, "Ignoring XID %lx (our xid is %lx)", (unsigned long) packet.xid, xid);
				continue;
			}
			
			if ((message = get_option(&packet, DHCP_MESSAGE_TYPE)) == NULL) {
				DEBUG(LOG_ERR, "couldnt get option from packet -- ignoring");
				continue;
			}
			
			switch (state) {
			case INIT_SELECTING:
				/* Must be a DHCPOFFER to one of our xid's */
				if (*message == DHCPOFFER) {
					if ((temp = get_option(&packet, DHCP_SERVER_ID))) {
						memcpy(&server_addr, temp, 4);
						xid = packet.xid;
						requested_ip = packet.yiaddr;
						*ipserver = server_addr;
						temp_addr.s_addr = server_addr;
						LOG(LOG_INFO, "Offer from %s ... waiting request ", inet_ntoa(temp_addr));
						
						/* enter requesting state */
						state = REQUESTING;
						timeout = time(0);
						packet_num = 0;
					} else {
						DEBUG(LOG_ERR, "No server ID in message");
					}
				}
				break;
			case REQUESTING:
				if (*message == DHCPACK) {
					if (!(temp = get_option(&packet, DHCP_LEASE_TIME))) {
						LOG(LOG_ERR, "No lease time with ACK, using 1 hour lease");
						*lease = 60*60;
					} else {
						memcpy(lease, temp, 4);
						*lease = ntohl(*lease);
					}
					if (!(temp = get_option(&packet, DHCP_DNS_SERVER))) {
					  LOG(LOG_ERR, "No dns server with ACK");
					} else {
					  if (*msdns == NULL) {
					    unsigned char *option = temp;
					    char *ptr;
					    int len = temp[OPT_LEN - 2];
					    *msdns = (char*) malloc(256*sizeof(char));
					    ptr = *msdns;

					    for(;;) {
					      ptr += sprintf(ptr, "ms-dns %d.%d.%d.%d",
							     option[0], option[1],
							     option[2], option[3]);
					      option += 4;
					      len -= 4;
					      
					      if (len <= 0) break;
					      *(ptr++) = ' ';
					    }
					  }
					  
					}
					if (!(temp = get_option(&packet, DHCP_WINS_SERVER))) {
					  LOG(LOG_ERR, "No wins server with ACK");
					} else {
					  if (*mswins == NULL) {
					    unsigned char *option = temp;
					    char *ptr;
					    int len = temp[OPT_LEN - 2];
					    *mswins = (char*) malloc(256*sizeof(char));
					    ptr = *mswins;

					    for(;;) {
					      ptr += sprintf(ptr, "ms-wins %d.%d.%d.%d",
							     option[0], option[1],
							     option[2], option[3]);
					      option += 4;
					      len -= 4;
					      
					      if (len <= 0) break;
					      *(ptr++) = ' ';
					    }
					  }
					  
					}
					temp_addr.s_addr = packet.yiaddr;
					LOG(LOG_INFO, "Lease of %s obtained  , lease time %ld", inet_ntoa(temp_addr),*lease);
					requested_ip = packet.yiaddr;
					if ( istoconfig ) run_script(&packet,"bound",ETHNAME,SCRIPTFILE);
					close(fd);
					*ipac = packet.yiaddr;
					return 1;
					
				} else if (*message == DHCPNAK) {
					LOG(LOG_INFO, "Received DHCP NAK");
					return 0;
				}
				break;
			}					
		} else {
			/* An error occured */
			DEBUG(LOG_ERR, "Error on select");
		}
		
	}
	return 0;
}

                                                                                                                            
void ReleaseIPtoDHCP(char *interf,unsigned long int  ipac,unsigned long int  ipserver,char *clientid) {
        send_release(ethaddr,ipserver, ipac,arp,clientid); 
        if ( interf ) run_script(NULL, "deconfig",ETHNAME,SCRIPTFILE);
}          

int RenewLease(unsigned long int  ipac,unsigned long int  ipserver,long int *newlease,char *clientid) {
	int fd,retval;
	unsigned long int xid;
	unsigned long timeout=0;
	int packet_num=0;
	char *temp,*message;
	fd_set rfds;
	struct dhcpMessage packet;
	struct timeval tv;

	if ((fd = listen_socket(INADDR_ANY, CLIENT_PORT,ETHNAME)) < 0) {
			LOG(LOG_ERR, "couldn't create server socket -- au revoir");
				return 0;
	}
	

	xid = random_xid();
	for(;;) {
		tv.tv_sec = timeout - time(0);
		tv.tv_usec = 0;
		FD_ZERO(&rfds);
		FD_SET(fd, &rfds);
		
		if (tv.tv_sec > 0) {
			retval = select(fd + 1, &rfds, NULL, NULL, &tv);
		} else retval = 0; 
		
		if (retval == 0) {
			if (packet_num < 3) {
				send_renew(ethaddr,xid, ipserver, ipac,ifindex,arp,clientid); 
				timeout = time(0) + ((packet_num == 2) ? 10 : 2);
				packet_num++;
			}
			else {
				close(fd);
				return 0;
			}
		}
		else if (retval > 0 && FD_ISSET(fd, &rfds)) {
		
			if (get_packet(&packet, fd) < 0) continue;	
			if (packet.xid != xid) {
				DEBUG(LOG_INFO, "Ignoring XID %lx (our xid is %lx)",(unsigned long) packet.xid, xid);
				continue;
			}		
			if ((message = get_option(&packet, DHCP_MESSAGE_TYPE)) == NULL) {
				DEBUG(LOG_ERR, "couldnt get option from packet -- ignoring");
				continue;
			}
			if (*message == DHCPACK) {
				if (!(temp = get_option(&packet, DHCP_LEASE_TIME))) {
					LOG(LOG_ERR, "No lease time with ACK, using 1 hour lease");
					*newlease= 60*60;
				} else {
					memcpy(newlease, temp, 4);
					*newlease= ntohl(*newlease);
					DEBUG(LOG_INFO, "New lease time with ACK: %ld",*newlease);
				}
				return 1;
			}
			else {
				close(fd);
				return 0;
			}


		}
		else {
			close(fd);
			return 0;
		}
	}


	return 0;
}













